#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import sys

from blint.analysis import report, start

blint_logo = """
██████╗ ██╗     ██╗███╗   ██╗████████╗
██╔══██╗██║     ██║████╗  ██║╚══██╔══╝
██████╔╝██║     ██║██╔██╗ ██║   ██║
██╔══██╗██║     ██║██║╚██╗██║   ██║
██████╔╝███████╗██║██║ ╚████║   ██║
╚═════╝ ╚══════╝╚═╝╚═╝  ╚═══╝   ╚═╝
"""


def build_args():
    """
    Constructs command line arguments for the blint tool
    """
    parser = argparse.ArgumentParser(
        prog="blint",
        description="Linting tool for binary files powered by lief.",
    )
    parser.add_argument(
        "-i",
        "--src",
        dest="src_dir_image",
        action="extend",
        nargs="+",
        help="Source directories, container images or binary files. Defaults to current directory.",
    )
    parser.add_argument(
        "-o",
        "--reports",
        dest="reports_dir",
        help="Reports directory. Defaults to reports.",
    )
    parser.add_argument(
        "--no-error",
        action="store_true",
        default=False,
        dest="noerror",
        help="Continue on error to prevent build from breaking",
    )
    parser.add_argument(
        "--no-banner",
        action="store_true",
        default=False,
        dest="no_banner",
        help="Do not display banner",
    )
    parser.add_argument(
        "--no-reviews",
        action="store_true",
        default=False,
        dest="no_reviews",
        help="Do not perform method reviews",
    )
    parser.add_argument(
        "--suggest-fuzzable",
        action="store_true",
        default=False,
        dest="suggest_fuzzable",
        help="Suggest functions and symbols for fuzzing based on a dictionary",
    )
    return parser.parse_args()


def parse_input(src):
    path = src[0]
    result = path.split("\n")
    result.pop()
    return result


def main():
    args = build_args()
    if not args.no_banner:
        print(blint_logo)
    src_dir = args.src_dir_image
    if os.getenv("IS_GHA"):
        src_dir = parse_input(args.src_dir_image)
    if args.reports_dir:
        reports_dir = args.reports_dir
    elif not src_dir:
        reports_base_dir, src_dir= os.getcwd(), [os.getcwd()]
        reports_dir = os.path.join(reports_base_dir, "reports")
    elif len(src_dir) == 1:
        reports_dir = os.path.dirname(src_dir[0])
    else:
        print("You must use the -o option to specify a reports output directory when scanning multiple sources.")
        exit()
    for dir in src_dir:
        if not os.path.exists(dir):
            print(f"{dir} is an invalid file or directory!")
            return
    # Create reports directory
    if reports_dir and not os.path.exists(reports_dir):
        os.makedirs(reports_dir)
    findings, reviews, files, fuzzables = start(args, src_dir, reports_dir)
    report(args, src_dir, reports_dir, findings, reviews, files, fuzzables)

    if os.getenv("IS_GHA"):
        if len(findings) > 0:
            sys.exit(1)


if __name__ == "__main__":
    main()



import logging
import os

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

custom_theme = Theme({"info": "cyan", "warning": "purple4", "danger": "bold red"})
console = Console(
    log_time=False,
    log_path=False,
    theme=custom_theme,
    width=280,
    color_system="auto",
    record=True,
)

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[
        RichHandler(
            console=console, markup=True, show_path=False, enable_link_path=False
        )
    ],
)
LOG = logging.getLogger(__name__)

# Set logging level
if os.getenv("SCAN_DEBUG_MODE") == "debug":
    LOG.setLevel(logging.DEBUG)

DEBUG = logging.DEBUG



import importlib
import json
import os
import re
import sys
import uuid
from collections import defaultdict
from datetime import datetime
from itertools import islice
from pathlib import Path

import yaml
from rich import box
from rich.progress import Progress
from rich.table import Table
from rich.terminal_theme import MONOKAI

from blint.binary import parse
from blint.logger import LOG, console
from blint.utils import find_exe_files, is_fuzzable_name, parse_pe_manifest

try:
    import importlib.resources

    # Defeat lazy module importers.
    importlib.resources.open_text
    HAVE_RESOURCE_READER = True
except ImportError:
    HAVE_RESOURCE_READER = False

review_files = []
if HAVE_RESOURCE_READER:
    try:
        review_methods_dir = importlib.resources.contents("blint.data.annotations")
        review_files = [rf for rf in review_methods_dir if rf.endswith(".yml")]
    except Exception:
        pass
else:
    review_methods_dir = Path(__file__).parent / "data" / "annotations"
    review_files = [p.as_posix() for p in Path(review_methods_dir).rglob("*.yml")]

rules_dict = {}
review_exe_dict = defaultdict(list)
review_methods_dict = defaultdict(list)
review_symbols_dict = defaultdict(list)
review_imports_dict = defaultdict(list)
review_entries_dict = defaultdict(list)
review_rules_cache = {}

# Debug mode
DEBUG_MODE = os.getenv("SCAN_DEBUG_MODE") == "debug"

# No of evidences per category
EVIDENCE_LIMIT = 5


def get_resource(package, resource):
    """Return a file handle on a named resource in a Package."""

    # Prefer ResourceReader APIs, as they are newest.
    if HAVE_RESOURCE_READER:
        # If we're in the context of a module, we could also use
        # ``__loader__.get_resource_reader(__name__).open_resource(resource)``.
        # We use open_binary() because it is simple.
        return importlib.resources.open_text(package, resource)

    # Fall back to __file__.

    # We need to first import the package so we can find its location.
    # This could raise an exception!
    mod = importlib.import_module(package)

    # Undefined __file__ will raise NameError on variable access.
    try:
        package_path = os.path.abspath(os.path.dirname(mod.__file__))
    except NameError:
        package_path = None

    if package_path is not None:
        # Warning: there is a path traversal attack possible here if
        # resource contains values like ../../../../etc/password. Input
        # must be trusted or sanitized before blindly opening files or
        # you may have a security vulnerability!
        resource_path = os.path.join(package_path, resource)

        return open(resource_path)

    # Could not resolve package path from __file__.
    LOG.warn("Unable to load resource: %s:%s" % (package, resource))
    return None


# Load the rules
with get_resource("blint.data", "rules.yml") as fp:
    raw_data = fp.read().split("---")
    for tmp_data in raw_data:
        if not tmp_data:
            continue
        rules_list = yaml.safe_load(tmp_data)
        for rule in rules_list:
            rules_dict[rule.get("id")] = rule

# Load the default review methods
for review_methods_file in review_files:
    if DEBUG_MODE:
        LOG.debug(f"Loading review file {review_methods_file}")
    with get_resource("blint.data.annotations", review_methods_file) as fp:
        raw_data = fp.read().split("---")
        for tmp_data in raw_data:
            if not tmp_data:
                continue
            methods_reviews_groups = yaml.safe_load(tmp_data)
            exe_type_list = methods_reviews_groups.get("exe_type")
            if isinstance(exe_type_list, str):
                exe_type_list = [exe_type_list]
            all_rules = methods_reviews_groups.get("rules")
            method_rules_dict = {}
            for rule in all_rules:
                method_rules_dict[rule.get("id")] = rule
                review_rules_cache[rule.get("id")] = rule
            for exe_type in exe_type_list:
                if methods_reviews_groups.get("group") == "METHOD_REVIEWS":
                    review_methods_dict[exe_type].append(method_rules_dict)
                elif methods_reviews_groups.get("group") == "EXE_REVIEWS":
                    review_exe_dict[exe_type].append(method_rules_dict)
                elif methods_reviews_groups.get("group") == "SYMBOL_REVIEWS":
                    review_symbols_dict[exe_type].append(method_rules_dict)
                elif methods_reviews_groups.get("group") == "IMPORT_REVIEWS":
                    review_imports_dict[exe_type].append(method_rules_dict)
                elif methods_reviews_groups.get("group") == "ENTRIES_REVIEWS":
                    review_entries_dict[exe_type].append(method_rules_dict)


def check_nx(f, metadata, rule_obj):
    if metadata.get("has_nx") is False:
        return False
    return True


def check_pie(f, metadata, rule_obj):
    if metadata.get("is_pie") is False:
        return False
    return True


def check_relro(f, metadata, rule_obj):
    if metadata.get("relro") == "no":
        return False
    return True


def check_canary(f, metadata, rule_obj):
    if metadata.get("has_canary") is False:
        return False
    return True


def check_rpath(f, metadata, rule_obj):
    # Do not recommend setting rpath or runpath
    if metadata.get("has_rpath") or metadata.get("has_runpath"):
        return False
    return True


def check_virtual_size(f, metadata, rule_obj):
    if metadata.get("virtual_size"):
        virtual_size = metadata.get("virtual_size") / 1024 / 1024
        size_limit = 30
        if rule_obj.get("limit"):
            limit = rule_obj.get("limit")
            limit = limit.replace("MB", "").replace("M", "")
            if isinstance(limit, str) and rule_obj.get("limit").isdigit():
                size_limit = int(rule_obj.get("limit"))
        return virtual_size < size_limit
    return True


def check_authenticode(f, metadata, rule_obj):
    if metadata.get("authenticode"):
        authenticode_obj = metadata.get("authenticode")
        vf = authenticode_obj.get("verification_flags", "").lower()
        if vf != "ok":
            return False
        if not authenticode_obj.get("cert_signer"):
            return False
        return True
    return True


def check_dll_characteristics(f, metadata, rule_obj):
    if metadata.get("dll_characteristics"):
        for c in rule_obj.get("mandatory_values", []):
            if c not in metadata.get("dll_characteristics"):
                return c
    return True


def check_codesign(f, metadata, rule_obj):
    if metadata.get("code_signature"):
        code_signature = metadata.get("code_signature")
        if code_signature and code_signature.get("available") is False:
            return False
        return True
    return True


def check_trust_info(f, metadata, rule_obj):
    if metadata.get("resources"):
        manifest = metadata.get("resources").get("manifest")
        if manifest:
            attribs_dict = parse_pe_manifest(manifest)
            if not attribs_dict:
                return True
            allowed_values = rule_obj.get("allowed_values", {})
            for k, v in allowed_values.items():
                manifest_k = attribs_dict.get(k)
                if isinstance(v, dict) and isinstance(manifest_k, dict):
                    for vk, vv in v.items():
                        if str(manifest_k.get(vk)).lower() != str(vv).lower():
                            return "{}:{}".format(vk, manifest_k.get(vk))
    return True


def run_checks(f, metadata):
    results = []
    if not rules_dict:
        LOG.warn("No rules loaded!")
        return None
    if not metadata:
        return None
    exe_type = metadata.get("exe_type")
    for cid, rule_obj in rules_dict.items():
        rule_exe_types = rule_obj.get("exe_types")
        # Skip rules that are not valid for this exe type
        if exe_type and rule_exe_types and exe_type not in rule_exe_types:
            continue
        cfn = getattr(sys.modules[__name__], cid.lower(), None)
        if cfn:
            result = cfn(f, metadata, rule_obj=rule_obj)
            if result is False or isinstance(result, str):
                aresult = {**rule_obj, "filename": f}
                if isinstance(result, str):
                    aresult["title"] = "{} ({})".format(aresult["title"], result)
                if metadata.get("name"):
                    aresult["exe_name"] = metadata.get("name")
                aresult["exe_type"] = exe_type
                results.append(aresult)
    return results


def run_review_methods_symbols(review_methods_list, functions_list):
    results = defaultdict(list)
    found_cid = defaultdict(int)
    found_pattern = defaultdict(int)
    found_function = {}
    for review_methods in review_methods_list:
        for cid, rule_obj in review_methods.items():
            if found_cid[cid] > EVIDENCE_LIMIT:
                continue
            patterns = rule_obj.get("patterns")
            for apattern in patterns:
                if (
                    found_pattern[apattern] > EVIDENCE_LIMIT
                    or found_cid[cid] > EVIDENCE_LIMIT
                ):
                    continue
                for afun in functions_list:
                    if (apattern.lower() in afun.lower()) and not found_function.get(
                        afun.lower()
                    ):
                        result = {
                            "pattern": apattern,
                            "function": afun,
                        }
                        results[cid].append(result)
                        found_cid[cid] += 1
                        found_pattern[apattern] += 1
                        found_function[afun.lower()] = True
    return results


def run_review(f, metadata):
    results = {}
    if not review_methods_dict:
        LOG.warn("No review methods loaded!")
        return None
    exe_type = metadata.get("exe_type")
    if not metadata or not exe_type:
        return None
    review_methods_list = review_methods_dict.get(exe_type)
    review_exe_list = review_exe_dict.get(exe_type)
    review_symbols_list = review_symbols_dict.get(exe_type)
    review_imports_list = review_imports_dict.get(exe_type)
    review_entries_list = review_entries_dict.get(exe_type)
    # Check if reviews are available for this exe type
    if (
        not review_methods_list
        and not review_exe_list
        and not review_symbols_list
        and not review_imports_list
        and not review_entries_list
    ):
        return None
    if review_methods_list or review_exe_list:
        functions_list = [
            re.sub(r"[*&()]", "", f.get("name", ""))
            for f in metadata.get("functions", [])
        ]
        if metadata.get("magic", "").startswith("PE"):
            functions_list += [
                f.get("name", "") for f in metadata.get("static_symbols", [])
            ]
        # If there are no function but static symbols use that instead
        if not functions_list and metadata.get("static_symbols"):
            functions_list = [
                f.get("name", "") for f in metadata.get("static_symbols", [])
            ]
        LOG.debug(f"Reviewing {len(functions_list)} functions")
        if review_methods_list:
            results.update(
                run_review_methods_symbols(review_methods_list, functions_list)
            )
        if review_exe_list:
            results.update(run_review_methods_symbols(review_exe_list, functions_list))
    if review_symbols_list or review_exe_list:
        symbols_list = [f.get("name", "") for f in metadata.get("dynamic_symbols", [])]
        symbols_list += [f.get("name", "") for f in metadata.get("static_symbols", [])]
        LOG.debug(f"Reviewing {len(symbols_list)} symbols")
        if review_symbols_list:
            results.update(
                run_review_methods_symbols(review_symbols_list, symbols_list)
            )
        if review_exe_list:
            results.update(run_review_methods_symbols(review_exe_list, symbols_list))
    if review_imports_list:
        imports_list = [f.get("name", "") for f in metadata.get("imports", [])]
        LOG.debug(f"Reviewing {len(imports_list)} imports")
        results.update(run_review_methods_symbols(review_imports_list, imports_list))
    if review_entries_list:
        entries_list = [
            f.get("name", "")
            for f in metadata.get("dynamic_entries", [])
            if f.get("tag") == "NEEDED"
        ]
        LOG.debug(f"Reviewing {len(entries_list)} dynamic entries")
        results.update(run_review_methods_symbols(review_entries_list, entries_list))
    return results


def run_prefuzz(f, metadata):
    functions_list = [
        {
            "name": re.sub(r"[*&()]", "", f.get("name", "")),
            "address": f.get("address", ""),
        }
        for f in metadata.get("functions", [])
    ]
    functions_list += [
        {
            "name": re.sub(r"[*&()]", "", f.get("name", "")),
            "address": f.get("address", ""),
        }
        for f in metadata.get("ctor_functions", [])
    ]
    functions_list += [
        {
            "name": re.sub(r"[*&()]", "", f.get("name", "")),
            "address": f.get("address", ""),
        }
        for f in metadata.get("exception_functions", [])
    ]
    functions_list += [
        {
            "name": re.sub(r"[*&()]", "", f.get("name", "")),
            "address": f.get("address", ""),
        }
        for f in metadata.get("unwind_functions", [])
    ]
    functions_list += [
        {"name": f.get("name", ""), "address": f.get("address", "")}
        for f in metadata.get("exports", [])
    ]
    fuzzables = [
        {"name": f.get("name"), "address": f.get("address", "").strip()}
        for f in functions_list
        if is_fuzzable_name(f.get("name"))
    ]
    LOG.debug(f"Found {len(fuzzables)} fuzzable methods")
    return fuzzables


def start(args, src, reports_dir):
    files = src
    findings = []
    reviews = []
    fuzzables = []
    for i in src:
        files += find_exe_files(i)
    with Progress(
        transient=True,
        redirect_stderr=True,
        redirect_stdout=True,
        refresh_per_second=1,
    ) as progress:
        task = progress.add_task(
            f"[green] Blinting {len(files)} binaries",
            total=len(files),
            start=True,
        )
        for f in files:
            progress.update(task, description=f"Processing [bold]{f}[/bold]")
            metadata = parse(f)
            exe_name = metadata.get("name", "")
            # Store raw metadata
            metadata_file = Path(reports_dir) / (
                os.path.basename(exe_name) + "-metadata.json"
            )
            LOG.debug(f"Metadata written to {metadata_file}")
            with open(metadata_file, mode="w") as ffp:
                json.dump(metadata, ffp, indent=True)
            progress.update(
                task, description=f"Checking [bold]{f}[/bold] against rules"
            )
            # Check security properties
            finding = run_checks(f, metadata)
            if finding:
                findings += finding
            # Perform symbol reviews
            if not args.no_reviews:
                progress.update(
                    task, description="Checking methods against review rules"
                )
                review = run_review(f, metadata)
                if review:
                    for cid, evidence in review.items():
                        aresult = {
                            **review_rules_cache.get(cid),
                            "evidence": evidence,
                            "filename": f,
                            "exe_name": exe_name,
                        }
                        del aresult["patterns"]
                        reviews.append(aresult)
            # Suggest fuzzable targets
            if args.suggest_fuzzable:
                fuzzdata = run_prefuzz(f, metadata)
                if fuzzdata:
                    fuzzables.append(
                        {"filename": f, "exe_name": exe_name, "methods": fuzzdata}
                    )
            progress.advance(task)
    return findings, reviews, files, fuzzables


def print_findings_table(findings, files):
    table = Table(
        title="BLint Findings",
        box=box.DOUBLE_EDGE,
        header_style="bold magenta",
        show_lines=True,
    )
    table.add_column("ID")
    if len(files) > 1:
        table.add_column("Binary")
    table.add_column("Title")
    table.add_column("Severity")
    for f in findings:
        severity = f.get("severity").upper()
        severity_fmt = "{}{}".format(
            "[bright_red]" if severity in ("CRITICAL", "HIGH") else "", severity
        )
        if len(files) > 1:
            table.add_row(
                f.get("id"),
                f.get("exe_name"),
                f.get("title"),
                severity_fmt,
            )
        else:
            table.add_row(
                f.get("id"),
                f.get("title"),
                severity_fmt,
            )
    console.print(table)


def print_reviews_table(reviews, files):
    table = Table(
        title="BLint Capability Review",
        box=box.DOUBLE_EDGE,
        header_style="bold magenta",
        show_lines=True,
    )
    table.add_column("ID")
    if len(files) > 1:
        table.add_column("Binary")
    table.add_column("Capabilities")
    table.add_column("Evidence (Top 5)", overflow="fold")
    for r in reviews:
        evidences = [e.get("function") for e in r.get("evidence")]
        evidences = list(islice(evidences, EVIDENCE_LIMIT))
        if len(files) > 1:
            table.add_row(
                r.get("id"),
                r.get("exe_name"),
                r.get("summary"),
                "\n".join(evidences),
            )
        else:
            table.add_row(
                r.get("id"),
                r.get("summary"),
                "\n".join(evidences),
            )
    console.print(table)


def report(args, src_dir, reports_dir, findings, reviews, files, fuzzables):
    run_uuid = os.environ.get("SCAN_ID", str(uuid.uuid4()))
    common_metadata = {
        "scan_id": run_uuid,
        "created": f"{datetime.now():%Y-%m-%d %H:%M:%S%z}",
    }
    if findings:
        print_findings_table(findings, files)
        findings_file = Path(reports_dir) / "findings.json"
        LOG.info(f"Findings written to {findings_file}")
        with open(findings_file, mode="w") as ffp:
            json.dump({**common_metadata, "findings": findings}, ffp, indent=True)
    if reviews:
        print_reviews_table(reviews, files)
        reviews_file = Path(reports_dir) / "reviews.json"
        LOG.info(f"Review written to {reviews_file}")
        with open(reviews_file, mode="w") as rfp:
            json.dump({**common_metadata, "reviews": reviews}, rfp, indent=True)
    if fuzzables:
        fuzzables_file = Path(reports_dir) / "fuzzables.json"
        LOG.info(f"Fuzzables data written to {fuzzables_file}")
        with open(fuzzables_file, mode="w") as rfp:
            json.dump({**common_metadata, "fuzzables": fuzzables}, rfp, indent=True)
    else:
        LOG.debug("No suggestion available for fuzzing")

    if not findings and not reviews:
        LOG.info(f":white_heavy_check_mark: No issues found in {src_dir}!")
    # Try console output as html
    html_file = Path(reports_dir) / "blint-output.html"
    console.save_html(html_file, theme=MONOKAI)
    LOG.info(f"HTML report written to {html_file}")



import sys

import lief
from lief import ELF, PE, MachO

from blint.logger import LOG
from blint.utils import calculate_entropy, check_secret, decode_base64

MIN_ENTROPY = 0.39
MIN_LENGTH = 80

lief.logging.disable()

address_fmt = "0x{:<10x}"


def parse_desc(e):
    return "{:02x}".format(e)


def is_shared_library(parsed_obj):
    if not parsed_obj:
        return False
    if parsed_obj.format == lief.EXE_FORMATS.ELF:
        return parsed_obj.header.file_type == lief.ELF.E_TYPE.DYNAMIC
    elif parsed_obj.format == lief.EXE_FORMATS.PE:
        return parsed_obj.header.has_characteristic(lief.PE.HEADER_CHARACTERISTICS.DLL)
    elif parsed_obj.format == lief.EXE_FORMATS.MACHO:
        return parsed_obj.header.file_type == lief.MachO.FILE_TYPES.DYLIB
    return False


def parse_notes(parsed_obj):
    metadata = {"notes": []}
    notes = parsed_obj.notes
    if len(notes):
        for idx, note in enumerate(notes):
            description = note.description
            description_str = " ".join(map(parse_desc, description[:16]))
            if len(description) > 16:
                description_str += " ..."
            type_str = note.type_core if note.is_core else note.type
            type_str = str(type_str).split(".")[-1]
            note_details = note.details
            note_details_str = ""
            sdk_version = ""
            ndk_version = ""
            ndk_build_number = ""
            abi = ""
            version_str = ""
            if type(note_details) == lief.ELF.AndroidNote:
                sdk_version = note_details.sdk_version
                ndk_version = note_details.ndk_version
                ndk_build_number = note_details.ndk_build_number
            if type(note_details) == lief.ELF.NoteAbi:
                version = note_details.version
                abi = str(note_details.abi)
                version_str = "{:d}.{:d}.{:d}".format(
                    version[0], version[1], version[2]
                )
            if note.is_core:
                note_details_str = note.details
            metadata["notes"].append(
                {
                    "description": str(description_str),
                    "type": type_str,
                    "details": note_details_str,
                    "sdk_version": sdk_version,
                    "ndk_version": ndk_version,
                    "ndk_build_number": ndk_build_number,
                    "abi": abi,
                    "version": version_str,
                }
            )
    return metadata["notes"]


def parse_uuid(e):
    return "{:02x}".format(e)


def parse_relro(parsed_obj):
    bind_now = False
    now = False
    try:
        parsed_obj.get(lief.ELF.SEGMENT_TYPES.GNU_RELRO)
    except lief.not_found:
        return "no"
    try:
        dynamic_tags = parsed_obj.get(lief.ELF.DYNAMIC_TAGS.FLAGS)
        if dynamic_tags:
            bind_now = lief.ELF.DYNAMIC_FLAGS.BIND_NOW in dynamic_tags
    except lief.not_found:
        pass
    try:
        dynamic_tags = parsed_obj.get(lief.ELF.DYNAMIC_TAGS.FLAGS_1)
        if dynamic_tags:
            now = lief.ELF.DYNAMIC_FLAGS_1.NOW in dynamic_tags
    except lief.not_found:
        pass
    if bind_now or now:
        return "full"
    else:
        return "partial"


def parse_functions(functions):
    try:
        LOG.debug("Parsing functions")
        func_list = []
        for idx, f in enumerate(functions):
            if f.name and f.address:
                cleaned_name = f.name.replace("..", "::")
                func_list.append(
                    {
                        "index": idx,
                        "name": cleaned_name,
                        "address": address_fmt.format(f.address),
                    }
                )
        return func_list
    except lief.exception:
        return []


def parse_strings(parsed_obj):
    try:
        LOG.debug("Parsing strings")
        strings_list = []
        strings = parsed_obj.strings
        for s in strings:
            if s and "[]" not in s and "{}" not in s:
                entropy = calculate_entropy(s)
                secret_type = check_secret(s)
                if (
                    entropy and (entropy > MIN_ENTROPY or len(s) > MIN_LENGTH)
                ) or secret_type:
                    strings_list.append(
                        {
                            "value": decode_base64(s) if s.endswith("==") else s,
                            "entropy": entropy,
                            "secret_type": secret_type,
                        }
                    )
        return strings_list
    except lief.exception:
        return []


def parse_symbols(symbols):
    try:
        LOG.debug("Parsing symbols")
        symbols_list = []
        exe_type = None
        for symbol in symbols:
            symbol_version = symbol.symbol_version if symbol.has_version else ""
            is_imported = False
            is_exported = False
            if symbol.imported:
                is_imported = True
            if symbol.exported:
                is_exported = True
            try:
                symbol_name = symbol.demangled_name
            except Exception:
                symbol_name = symbol.name
            if not exe_type:
                exe_type = guess_exe_type(symbol_name)
            symbols_list.append(
                {
                    "name": symbol_name,
                    "type": str(symbol.type).split(".")[-1],
                    "value": symbol.value,
                    "visibility": str(symbol.visibility).split(".")[-1],
                    "binding": str(symbol.binding).split(".")[-1],
                    "is_imported": is_imported,
                    "is_exported": is_exported,
                    "information": symbol.information,
                    "is_function": symbol.is_function,
                    "is_static": symbol.is_static,
                    "is_variable": symbol.is_variable,
                    "version": str(symbol_version),
                }
            )
        return symbols_list, exe_type
    except lief.exception:
        return [], None


def parse_interpreter(parsed_obj):
    try:
        return parsed_obj.interpreter
    except lief.exception as e:
        return ""


def detect_exe_type(parsed_obj, metadata):
    try:
        if parsed_obj.has_section(".note.go.buildid"):
            return "gobinary"
        if (
            parsed_obj.has_section(".note.gnu.build-id")
            or "musl" in metadata.get("interpreter")
            or "ld-linux" in metadata.get("interpreter")
        ):
            return "genericbinary"
        if metadata.get("machine_type") and metadata.get("file_type"):
            return "{}-{}".format(
                metadata.get("machine_type"), metadata.get("file_type")
            ).lower()
        if metadata["relro"] in ("partial", "full"):
            return "genericbinary"
    except lief.exception:
        return ""


def guess_exe_type(symbol_name):
    exe_type = None
    if "golang" in symbol_name or "_cgo_" in symbol_name:
        exe_type = "gobinary"
    if "_rust_" in symbol_name:
        exe_type = "genericbinary"
    if "DotNetRuntimeInfo" in symbol_name:
        exe_type = "dotnetbinary"
    return exe_type


def parse_pe_data(parsed_obj):
    try:
        LOG.debug("Parsing data dictionaries")
        data_list = []
        data_directories = parsed_obj.data_directories
        for directory in data_directories:
            section_name = ""
            section_chars = ""
            section_entropy = ""
            dir_type = str(directory.type).split(".")[-1]
            if not dir_type.startswith("?") and directory.size:
                if directory.has_section:
                    if directory.section.has_characteristic:
                        section_chars = ", ".join(
                            [
                                str(chara).split(".")[-1]
                                for chara in directory.section.characteristics_lists
                            ]
                        )
                    section_name = directory.section.name
                    section_entropy = directory.section.entropy
                data_list.append(
                    {
                        "name": section_name,
                        "type": dir_type,
                        "rva": directory.rva,
                        "size": directory.size,
                        "section_characteristics": section_chars,
                        "section_entropy": section_entropy,
                    }
                )
        return data_list
    except lief.exception:
        return None


def process_pe_resources(parsed_obj):
    try:
        LOG.debug("Parsing PE resources")
        rm = parsed_obj.resources_manager
        metadata = {
            "has_accelerator": rm.has_accelerator,
            "has_dialogs": rm.has_dialogs,
            "has_html": rm.has_html,
            "has_icons": rm.has_icons,
            "has_manifest": rm.has_manifest,
            "has_string_table": rm.has_string_table,
            "has_version": rm.has_version,
            "html": rm.html if rm.has_html else None,
            "manifest": rm.manifest.replace("\\xef\\xbb\\xbf", "")
            if rm.has_manifest
            else None,
            "version_info": str(rm.version) if rm.has_version else None,
        }
        return metadata
    except lief.exception:
        return None


def process_pe_signature(parsed_obj):
    try:
        signature_list = []
        for idx, sig in enumerate(parsed_obj.signatures):
            ci = sig.content_info
            signature_obj = {
                "version": sig.version,
                "digest_algorithm": str(sig.digest_algorithm),
                "content_info": {
                    "content_type": lief.PE.oid_to_string(ci.content_type),
                    "digest_algorithm": str(ci.digest_algorithm),
                    "digest": ci.digest.hex(),
                },
            }
            signers_list = []
            for signer in sig.signers:
                signer_obj = {
                    "version": signer.version,
                    "serial_number": signer.serial_number.hex(),
                    "issuer": str(signer.issuer),
                    "digest_algorithm": str(signer.digest_algorithm),
                    "encryption_algorithm": str(signer.encryption_algorithm),
                    "encrypted_digest": signer.encrypted_digest.hex(),
                }
                signers_list.append(signer_obj)
            signature_obj["signers"] = signers_list
            signature_list.append(signature_obj)
        return signature_list
    except lief.exception:
        return None


def parse_pe_authenticode(parsed_obj):
    try:
        LOG.debug("Parsing authentihash")
        authenticode = {}
        sep = ":" if sys.version_info.minor > 7 else ()
        authenticode["md5_hash"] = parsed_obj.authentihash_md5.hex(*sep)
        authenticode["sha256_hash"] = parsed_obj.authentihash_sha256.hex(*sep)
        authenticode["sha512_hash"] = parsed_obj.authentihash_sha512.hex(*sep)
        authenticode["sha1_hash"] = parsed_obj.authentihash(
            lief.PE.ALGORITHMS.SHA_1
        ).hex(*sep)
        authenticode["verification_flags"] = str(parsed_obj.verify_signature()).replace(
            "VERIFICATION_FLAGS.", ""
        )
        if (
            parsed_obj.signatures
            and len(parsed_obj.signatures) > 0
            and parsed_obj.signatures[0].signers
        ):
            cert_signer_str = str(parsed_obj.signatures[0].signers[0].cert)
            cert_signer_obj = {}
            for p in cert_signer_str.split("\n"):
                tmpA = p.split(" : ")
                if len(tmpA) == 2:
                    tmpKey = tmpA[0].strip().replace(" ", "_")
                    if "version" in tmpKey:
                        tmpKey = "version"
                    cert_signer_obj[tmpKey] = tmpA[1].strip()
            authenticode["cert_signer"] = cert_signer_obj
        return authenticode
    except lief.exception:
        return None


def parse_pe_symbols(symbols):
    try:
        LOG.debug("Parsing symbols")
        symbols_list = []
        exe_type = None
        for symbol in symbols:
            if symbol.section_number <= 0:
                section_nb_str = str(
                    PE.SYMBOL_SECTION_NUMBER(symbol.section_number)
                ).split(".")[-1]
            else:
                try:
                    section_nb_str = symbol.section.name
                except Exception:
                    section_nb_str = "section<{:d}>".format(symbol.section_number)
            try:
                if not exe_type:
                    try:
                        exe_type = guess_exe_type(symbol.name.lower())
                    except Exception:
                        pass
                if symbol.name:
                    symbols_list.append(
                        {
                            "name": symbol.name.replace("..", "::"),
                            "value": symbol.value,
                            "id": section_nb_str,
                            "base_type": str(symbol.base_type).split(".")[-1],
                            "complex_type": str(symbol.complex_type).split(".")[-1],
                            "storage_class": str(symbol.storage_class).split(".")[-1],
                        }
                    )
            except Exception:
                pass
        return symbols_list, exe_type
    except lief.exception:
        return [], None


def parse_pe_imports(imports):
    try:
        LOG.debug("Parsing imports")
        imports_list = []
        dlls = set()
        for import_ in imports:
            entries = import_.entries
            for entry in entries:
                if entry.name:
                    dlls.add(import_.name)
                    imports_list.append(
                        {
                            "name": "{}::{}".format(import_.name, entry.name),
                            "short_name": entry.name,
                            "data": entry.data,
                            "iat_value": entry.iat_value,
                            "hint": entry.hint,
                        }
                    )
        dll_list = [{"name": d, "tag": "NEEDED"} for d in list(dlls)]
        return imports_list, dll_list
    except lief.exception:
        return []


def parse_pe_exports(exports):
    try:
        LOG.debug("Parsing exports")
        exports_list = []
        entries = exports.entries
        for entry in entries:
            extern = "[EXTERN]" if entry.is_extern else ""
            if entry.name:
                metadata = {
                    "name": entry.name,
                    "ordinal": entry.ordinal,
                    "address": address_fmt.format(entry.address),
                    "extern": extern,
                }
            fwd = entry.forward_information if entry.is_forwarded else None
            metadata["is_forwarded"] = entry.is_forwarded
            if fwd:
                metadata["fwd_library"] = fwd.library
                metadata["fwd_function"] = fwd.function
            exports_list.append(metadata)
        return exports_list
    except lief.exception:
        return []


def parse_macho_symbols(symbols):
    try:
        LOG.debug("Parsing symbols")
        symbols_list = []
        exe_type = None
        if len(symbols) == 0:
            return [], None
        for symbol in symbols:
            libname = ""
            if symbol.has_binding_info and symbol.binding_info.has_library:
                libname = symbol.binding_info.library.name

            symbol_value = (
                symbol.value
                if symbol.value > 0 or not symbol.has_binding_info
                else address_fmt.format(symbol.binding_info.address)
            )

            try:
                symbol_name = symbol.demangled_name
            except Exception:
                symbol_name = symbol.name
            symbol_name = symbol_name.replace("..", "::")
            if not exe_type:
                exe_type = guess_exe_type(symbol_name)
            symbols_list.append(
                {
                    "name": "{}::{}".format(libname, symbol_name)
                    if libname
                    else symbol_name,
                    "short_name": symbol_name,
                    "type": symbol.type,
                    "num_sections": symbol.numberof_sections,
                    "description": symbol.description,
                    "value": symbol_value,
                }
            )
        return symbols_list, exe_type
    except lief.exception:
        return [], None


def parse(exe_file):
    """
    Parse the executable using lief and capture the metadata

    :param: exe_file Binary file
    :return Metadata dict
    """
    metadata = {"file_path": exe_file}
    try:
        parsed_obj = lief.parse(exe_file)
        metadata["is_shared_library"] = is_shared_library(parsed_obj)
        # ELF Binary
        if isinstance(parsed_obj, ELF.Binary):
            metadata["binary_type"] = "ELF"
            header = parsed_obj.header
            identity = header.identity
            eflags_str = ""
            if header.machine_type == lief.ELF.ARCH.ARM:
                eflags_str = " - ".join(
                    [str(s).split(".")[-1] for s in header.arm_flags_list]
                )
            if header.machine_type in [
                lief.ELF.ARCH.MIPS,
                lief.ELF.ARCH.MIPS_RS3_LE,
                lief.ELF.ARCH.MIPS_X,
            ]:
                eflags_str = " - ".join(
                    [str(s).split(".")[-1] for s in header.mips_flags_list]
                )
            if header.machine_type == lief.ELF.ARCH.PPC64:
                eflags_str = " - ".join(
                    [str(s).split(".")[-1] for s in header.ppc64_flags_list]
                )
            if header.machine_type == lief.ELF.ARCH.HEXAGON:
                eflags_str = " - ".join(
                    [str(s).split(".")[-1] for s in header.hexagon_flags_list]
                )
            metadata["magic"] = ("{:<02x} " * 8).format(*identity[:8]).strip()
            metadata["class"] = str(header.identity_class).split(".")[-1]
            metadata["endianness"] = str(header.identity_data).split(".")[-1]
            metadata["identity_version"] = str(header.identity_version).split(".")[-1]
            metadata["identity_os_abi"] = str(header.identity_os_abi).split(".")[-1]
            metadata["identity_abi_version"] = header.identity_abi_version
            metadata["file_type"] = str(header.file_type).split(".")[-1]
            metadata["machine_type"] = str(header.machine_type).split(".")[-1]
            metadata["object_file_version"] = str(header.object_file_version).split(
                "."
            )[-1]
            metadata["entrypoint"] = header.entrypoint
            metadata["processor_flag"] = str(header.processor_flag) + eflags_str
            metadata["name"] = parsed_obj.name
            metadata["imagebase"] = parsed_obj.imagebase
            metadata["interpreter"] = parse_interpreter(parsed_obj)
            metadata["is_pie"] = parsed_obj.is_pie
            metadata["virtual_size"] = parsed_obj.virtual_size
            metadata["has_nx"] = parsed_obj.has_nx
            metadata["relro"] = parse_relro(parsed_obj)
            metadata["exe_type"] = detect_exe_type(parsed_obj, metadata)
            # Canary check
            canary_sections = ["__stack_chk_fail", "__intel_security_cookie"]
            for section in canary_sections:
                try:
                    if parsed_obj.get_symbol(section):
                        metadata["has_canary"] = True
                        break
                except lief.not_found:
                    metadata["has_canary"] = False
            # rpath check
            try:
                if parsed_obj.get(lief.ELF.DYNAMIC_TAGS.RPATH):
                    metadata["has_rpath"] = True
            except lief.not_found:
                metadata["has_rpath"] = False
            # runpath check
            try:
                if parsed_obj.get(lief.ELF.DYNAMIC_TAGS.RUNPATH):
                    metadata["has_runpath"] = True
            except lief.not_found:
                metadata["has_runpath"] = False
            static_symbols = parsed_obj.static_symbols
            if len(static_symbols):
                metadata["static"] = True
            dynamic_entries = parsed_obj.dynamic_entries
            if len(dynamic_entries):
                metadata["dynamic_entries"] = []
                for entry in dynamic_entries:
                    if entry.tag == ELF.DYNAMIC_TAGS.NULL:
                        continue
                    if entry.tag in [
                        ELF.DYNAMIC_TAGS.SONAME,
                        ELF.DYNAMIC_TAGS.NEEDED,
                        ELF.DYNAMIC_TAGS.RUNPATH,
                        ELF.DYNAMIC_TAGS.RPATH,
                    ]:
                        metadata["dynamic_entries"].append(
                            {
                                "name": entry.name,
                                "tag": str(entry.tag).split(".")[-1],
                                "value": entry.value,
                            }
                        )
                        # Detect dotnet binary
                        if "netcoredeps" in entry.name:
                            metadata["exe_type"] = "dotnetbinary"
            try:
                symbols_version = parsed_obj.symbols_version
                if len(symbols_version):
                    metadata["symbols_version"] = []
                    symbol_version_auxiliary_cache = {}
                    for entry in symbols_version:
                        symbol_version_auxiliary = entry.symbol_version_auxiliary
                        if (
                            symbol_version_auxiliary
                            and not symbol_version_auxiliary_cache.get(
                                symbol_version_auxiliary.name
                            )
                        ):
                            symbol_version_auxiliary_cache[
                                symbol_version_auxiliary.name
                            ] = True
                            metadata["symbols_version"].append(
                                {
                                    "name": symbol_version_auxiliary.name,
                                    "hash": symbol_version_auxiliary.hash,
                                    "value": entry.value,
                                }
                            )
            except lief.exception:
                metadata["symbols_version"] = []
            try:
                notes = parsed_obj.notes
                if notes:
                    metadata["notes"] = parse_notes(parsed_obj)
            except lief.exception:
                pass
            metadata["strings"] = parse_strings(parsed_obj)
            try:
                metadata["static_symbols"], exe_type = parse_symbols(
                    parsed_obj.static_symbols
                )
                if exe_type:
                    metadata["exe_type"] = exe_type
            except lief.exception:
                pass
            try:
                metadata["dynamic_symbols"], exe_type = parse_symbols(
                    parsed_obj.dynamic_symbols
                )
                if exe_type:
                    metadata["exe_type"] = exe_type
            except lief.exception:
                pass
            try:
                metadata["functions"] = parse_functions(parsed_obj.functions)
            except lief.exception:
                pass
            try:
                metadata["ctor_functions"] = parse_functions(parsed_obj.ctor_functions)
            except lief.exception:
                pass
        elif isinstance(parsed_obj, PE.Binary):
            # PE
            # Parse header
            try:
                metadata["binary_type"] = "PE"
                metadata["name"] = parsed_obj.name
                metadata["is_pie"] = parsed_obj.is_pie
                metadata["is_reproducible_build"] = parsed_obj.is_reproducible_build
                metadata["virtual_size"] = parsed_obj.virtual_size
                metadata["has_nx"] = parsed_obj.has_nx
                dos_header = parsed_obj.dos_header
                metadata["magic"] = str(dos_header.magic)
                header = parsed_obj.header
                optional_header = parsed_obj.optional_header
                metadata[
                    "used_bytes_in_the_last_page"
                ] = dos_header.used_bytes_in_the_last_page
                metadata["file_size_in_pages"] = dos_header.file_size_in_pages
                metadata["num_relocation"] = dos_header.numberof_relocation
                metadata[
                    "header_size_in_paragraphs"
                ] = dos_header.header_size_in_paragraphs
                metadata[
                    "minimum_extra_paragraphs"
                ] = dos_header.minimum_extra_paragraphs
                metadata[
                    "maximum_extra_paragraphs"
                ] = dos_header.maximum_extra_paragraphs
                metadata["initial_relative_ss"] = dos_header.initial_relative_ss
                metadata["initial_sp"] = dos_header.initial_sp
                metadata["checksum"] = dos_header.checksum
                metadata["initial_ip"] = dos_header.initial_ip
                metadata["initial_relative_cs"] = dos_header.initial_relative_cs
                metadata["address_relocation_table"] = address_fmt.format(
                    dos_header.addressof_relocation_table
                )
                metadata["overlay_number"] = dos_header.overlay_number
                metadata["oem_id"] = dos_header.oem_id
                metadata["oem_info"] = dos_header.oem_info
                metadata["address_new_exeheader"] = address_fmt.format(
                    dos_header.addressof_new_exeheader
                )
                metadata["characteristics"] = ", ".join(
                    [str(chara).split(".")[-1] for chara in header.characteristics_list]
                )
                metadata["num_sections"] = header.numberof_sections
                metadata["time_date_stamps"] = header.time_date_stamps
                metadata["pointer_symbol_table"] = header.pointerto_symbol_table
                metadata["num_symbols"] = header.numberof_symbols
                metadata["size_optional_header"] = header.sizeof_optional_header
                metadata["dll_characteristics"] = ", ".join(
                    [
                        str(chara).split(".")[-1]
                        for chara in optional_header.dll_characteristics_lists
                    ]
                )
                metadata["subsystem"] = str(optional_header.subsystem).split(".")[-1]
                metadata["is_gui"] = (
                    True if metadata["subsystem"] == "WINDOWS_GUI" else False
                )
                metadata["exe_type"] = (
                    "PE32" if optional_header.magic == PE.PE_TYPE.PE32 else "PE64"
                )
                metadata["major_linker_version"] = optional_header.major_linker_version
                metadata["minor_linker_version"] = optional_header.minor_linker_version
                metadata["sizeof_code"] = optional_header.sizeof_code
                metadata[
                    "sizeof_initialized_data"
                ] = optional_header.sizeof_initialized_data
                metadata[
                    "sizeof_uninitialized_data"
                ] = optional_header.sizeof_uninitialized_data
                metadata["addressof_entrypoint"] = address_fmt.format(
                    optional_header.addressof_entrypoint
                )
                metadata["baseof_code"] = optional_header.baseof_code
                metadata["baseof_data"] = optional_header.baseof_data
                metadata["imagebase"] = optional_header.imagebase
                metadata["section_alignment"] = optional_header.section_alignment
                metadata["file_alignment"] = optional_header.file_alignment
                metadata[
                    "major_operating_system_version"
                ] = optional_header.major_operating_system_version
                metadata[
                    "minor_operating_system_version"
                ] = optional_header.minor_operating_system_version
                metadata["major_image_version"] = optional_header.major_image_version
                metadata["minor_image_version"] = optional_header.minor_image_version
                metadata[
                    "major_subsystem_version"
                ] = optional_header.major_subsystem_version
                metadata[
                    "minor_subsystem_version"
                ] = optional_header.minor_subsystem_version
                metadata["win32_version_value"] = optional_header.win32_version_value
                metadata["sizeof_image"] = optional_header.sizeof_image
                metadata["sizeof_headers"] = optional_header.sizeof_headers
                metadata["checksum"] = optional_header.checksum
                metadata["sizeof_stack_reserve"] = optional_header.sizeof_stack_reserve
                metadata["sizeof_stack_commit"] = optional_header.sizeof_stack_commit
                metadata["sizeof_heap_reserve"] = optional_header.sizeof_heap_reserve
                metadata["sizeof_heap_commit"] = optional_header.sizeof_heap_commit
                metadata["loader_flags"] = optional_header.loader_flags
                metadata[
                    "numberof_rva_and_size"
                ] = optional_header.numberof_rva_and_size
            except lief.exception:
                pass
            try:
                metadata["data_directories"] = parse_pe_data(parsed_obj)
            except lief.exception:
                pass
            try:
                metadata["authenticode"] = parse_pe_authenticode(parsed_obj)
            except lief.exception:
                pass
            try:
                metadata["signatures"] = process_pe_signature(parsed_obj)
            except lief.exception:
                pass
            try:
                metadata["resources"] = process_pe_resources(parsed_obj)
            except lief.exception:
                pass
            try:
                metadata["static_symbols"], exe_type = parse_pe_symbols(
                    parsed_obj.symbols
                )
                if exe_type:
                    metadata["exe_type"] = exe_type
            except lief.exception:
                pass
            try:
                metadata["imports"], metadata["dynamic_entries"] = parse_pe_imports(
                    parsed_obj.imports
                )
            except lief.exception:
                pass
            try:
                metadata["exports"] = parse_pe_exports(parsed_obj.get_export())
            except lief.exception:
                pass
            try:
                metadata["functions"] = parse_functions(parsed_obj.functions)
            except lief.exception:
                pass
            try:
                metadata["ctor_functions"] = parse_functions(parsed_obj.ctor_functions)
            except lief.exception:
                pass
            try:
                metadata["exception_functions"] = parse_functions(
                    parsed_obj.exception_functions
                )
            except lief.exception:
                pass
            try:
                tls = parsed_obj.tls
                if tls and tls.sizeof_zero_fill:
                    metadata["tls_address_index"] = tls.addressof_index
                    metadata["tls_sizeof_zero_fill"] = tls.sizeof_zero_fill
                    metadata["tls_data_template_len"] = len(tls.data_template)
                    metadata["tls_characteristics"] = tls.characteristics
                    metadata["tls_section_name"] = tls.section.name
                    metadata["tls_directory_type"] = str(tls.directory.type)
            except lief.exception:
                pass
        elif isinstance(parsed_obj, MachO.Binary):
            # MachO
            metadata["binary_type"] = "MachO"
            metadata["name"] = parsed_obj.name
            metadata["imagebase"] = parsed_obj.imagebase
            metadata["is_pie"] = parsed_obj.is_pie
            metadata["has_nx"] = parsed_obj.has_nx
            metadata["exe_type"] = "MachO"
            try:
                version = (
                    parsed_obj.version_min.version if parsed_obj.version_min else None
                )
                sdk = parsed_obj.version_min.sdk if parsed_obj.version_min else None
                source_version = (
                    parsed_obj.source_version.version
                    if parsed_obj.source_version
                    else None
                )
                if source_version:
                    metadata["source_version"] = "{:d}.{:d}.{:d}.{:d}.{:d}".format(
                        *source_version
                    )
                if version:
                    metadata["version"] = "{:d}.{:d}.{:d}".format(*version)
                if sdk:
                    metadata["sdk"] = "{:d}.{:d}.{:d}".format(*sdk)
            except lief.exception:
                pass
            try:
                build_version = parsed_obj.build_version
                if build_version:
                    metadata["platform"] = str(build_version.platform).split(".")[-1]
                    metadata["minos"] = "{:d}.{:d}.{:d}".format(*build_version.minos)
                    metadata["sdk"] = "{:d}.{:d}.{:d}".format(*build_version.sdk)
                    tools = build_version.tools
                    if len(tools) > 0:
                        metadata["tools"] = []
                        for tool in tools:
                            tool_str = str(tool.tool).split(".")[-1]
                            metadata["tools"].append(
                                {
                                    "tool": tool_str,
                                    "version": "{}.{}.{}".format(*tool.version),
                                }
                            )
            except lief.exception:
                pass
            try:
                if parsed_obj.has_encryption_info:
                    encryption_info = parsed_obj.encryption_info
                    if encryption_info:
                        metadata["encryption_info"] = {
                            "crypt_offset": encryption_info.crypt_offset,
                            "crypt_size": encryption_info.crypt_size,
                            "crypt_id": encryption_info.crypt_id,
                        }
            except lief.exception:
                pass
            try:
                sinfo = parsed_obj.sub_framework
                if sinfo:
                    metadata["umbrella"] = sinfo.umbrella
            except lief.exception:
                pass
            try:
                cmd = parsed_obj.rpath
                if cmd:
                    metadata["has_rpath"] = True
                    metadata["rpath"] = cmd.path
            except lief.exception:
                metadata["has_rpath"] = False
            try:
                cmd = parsed_obj.uuid
                if cmd:
                    uuid_str = " ".join(map(parse_uuid, cmd.uuid))
                    metadata["uuid"] = str(uuid_str)
            except lief.exception:
                pass
            try:
                if parsed_obj.libraries:
                    metadata["libraries"] = []
                    for library in parsed_obj.libraries:
                        current_version_str = "{:d}.{:d}.{:d}".format(
                            *library.current_version
                        )
                        compatibility_version_str = "{:d}.{:d}.{:d}".format(
                            *library.compatibility_version
                        )
                        metadata["libraries"].append(
                            {
                                "name": library.name,
                                "timestamp": library.timestamp,
                                "version": current_version_str,
                                "compatibility_version": compatibility_version_str,
                            }
                        )
            except lief.exception:
                pass
            try:
                header = parsed_obj.header
                flags_str = ", ".join(
                    [str(s).split(".")[-1] for s in header.flags_list]
                )
                metadata["magic"] = str(header.magic).split(".")[-1]
                metadata["cpu_type"] = str(header.cpu_type).split(".")[-1]
                metadata["cpu_subtype"] = header.cpu_subtype
                metadata["file_type"] = str(header.file_type).split(".")[-1]
                metadata["flags"] = flags_str
                metadata["number_commands"] = header.nb_cmds
                metadata["size_commands"] = header.sizeof_cmds
                metadata["reserved"] = header.reserved
            except lief.exception:
                pass
            try:
                if parsed_obj.main_command:
                    metadata["has_main_command"] = True
                if parsed_obj.thread_command:
                    metadata["has_thread_command"] = True
            except lief.not_found:
                metadata["has_main"] = False
                metadata["has_thread_command"] = False
            try:
                metadata["functions"] = parse_functions(parsed_obj.functions)
            except lief.exception:
                pass
            try:
                metadata["ctor_functions"] = parse_functions(parsed_obj.ctor_functions)
            except lief.exception:
                pass
            try:
                metadata["unwind_functions"] = parse_functions(
                    parsed_obj.unwind_functions
                )
            except lief.exception:
                pass
            try:
                metadata["static_symbols"], exe_type = parse_macho_symbols(
                    parsed_obj.symbols
                )
                if exe_type:
                    metadata["exe_type"] = exe_type
            except lief.exception:
                pass
            try:
                if parsed_obj.dylinker:
                    metadata["dylinker"] = parsed_obj.dylinker.name
            except lief.exception:
                pass
            try:
                if parsed_obj.has_code_signature:
                    code_signature = parsed_obj.code_signature
                    metadata["code_signature"] = {
                        "available": True if code_signature.size else False,
                        "data": str(bytes(code_signature.data).hex()),
                        "data_size": str(code_signature.data_size),
                        "size": str(code_signature.size),
                    }
                if (
                    not parsed_obj.has_code_signature
                    and parsed_obj.has_code_signature_dir
                ):
                    code_signature = parsed_obj.has_code_signature_dir
                    metadata["code_signature"] = {
                        "available": True if code_signature.size else False,
                        "data": str(bytes(code_signature.data).hex()),
                        "data_size": str(code_signature.data_size),
                        "size": str(code_signature.size),
                    }
                if (
                    not parsed_obj.has_code_signature
                    and not parsed_obj.has_code_signature_dir
                ):
                    metadata["code_signature"] = {"available": False}
                if parsed_obj.has_data_in_code:
                    data_in_code = parsed_obj.data_in_code
                    metadata["data_in_code"] = {
                        "data": str(bytes(data_in_code.data).hex()),
                        "data_size": str(data_in_code.data_size),
                        "size": str(data_in_code.size),
                    }
            except lief.exception:
                pass
    except lief.exception as e:
        LOG.exception(e)
    return metadata



import base64
import math
import os
import re
import string
from pathlib import Path

from defusedxml.ElementTree import fromstring

charset = string.digits + string.ascii_letters + r"""!&@"""

# Default ignore list
ignore_directories = [
    ".git",
    ".svn",
    ".mvn",
    ".idea",
    "backup",
    "docs",
    "tests",
    "test",
    "report",
    "reports",
    "node_modules",
    ".terraform",
    ".serverless",
    "venv",
    "examples",
    "tutorials",
    "samples",
    "migrations",
    "db_migrations",
    "unittests",
    "unittests_legacy",
    "stubs",
    "mock",
    "mocks",
]

ignore_files = [
    ".pyc",
    ".gz",
    ".tar",
    ".tar.gz",
    ".tar",
    ".log",
    ".tmp",
    ".gif",
    ".png",
    ".jpg",
    ".webp",
    ".webm",
    ".icns",
    ".pcm",
    ".wav",
    ".mp3",
    ".pdf",
    ".doc",
    ".docx",
    ".xls",
    ".xlsx",
    ".d.ts",
    ".min.js",
    ".min.css",
    ".eslintrc.js",
    ".babelrc.js",
    ".spec.js",
    ".spec.ts",
    ".component.spec.js",
    ".component.spec.ts",
    ".data.js",
    ".data.ts",
    ".bundle.js",
    ".snap",
    ".pb.go",
    ".tests.py",
    ".vdb",
    ".txt",
    ".strings",
    ".nib",
]

strings_allowlist = [
    "()",
    "[]",
    "{}",
    " ",
    "&*",
    "(*",
    "$*",
    "!*",
    "*func",
    "*map",
    "Enabled",
    "crypto/",
    "readBase",
    "toFloat",
    "encoding/",
    "*tls",
    "*http",
    "*grpc",
    "protobuf:",
    "*runtime",
    "*x509",
    "*[",
    "*struct",
    "github.com",
    "vendor/",
    "golang.org",
    "gopkg.in",
    "*reflect",
    "*base64",
    "*pgtype",
    "*dsa",
    "*log",
    "*sql",
    "*zip",
    "*json",
    "*yaml",
    "*xz",
    "*errors",
    "*flag",
    "*object",
    "*ssh",
    "*syntax",
    "*zip",
    "json:",
    "basic_string",
    "std::",
    "vector::",
    "coreclr_",
    "deps_resolver_",
    "deps_json_",
    "NativeExceptionHolder",
    "System.Runtime",
    "Microsoft-",
    "ProfilerEnum",
    "FastSerialization",
    "InlineDiscretionary",
    "src/libraries",
    "ECDHE-ECDSA-AES256-GCM",
    "setsockopt",
    ".jar",
]

# Method names containing common verbs that could be fuzzed
fuzzable_names = [
    "create",
    "delete",
    "update",
    "insert",
    "upsert",
    "post",
    "put",
    "get",
    "index",
    "encrypt",
    "decrypt",
    "load",
    "import",
    "export",
    "encode",
    "decode",
    "obfuscate",
    "store",
    "fuzz",
    "route",
    "search",
    "find",
    "open",
    "send",
    "receive",
    "approve",
    "reject",
    "password",
    "lock",
    "find",
    "allocate",
    "peer",
    "block",
    "socket",
    "database",
    "remote",
    "api",
    "build",
    "transact",
    "engage",
    "quote",
    "unicode",
    "escape",
    "invoke",
    "execute",
    "process",
    "shell",
    "env",
    "valid",
    "saniti",
    "check",
    "audit",
    "follow",
    "link",
]

# import nltk
# nltk.download('treebank')
# nltk.download('universal_tagset')
# wsj = nltk.corpus.treebank.tagged_words(tagset='universal')
# word_tag_fd = nltk.FreqDist(wsj)
# [wt[0] for (wt, _) in word_tag_fd.most_common() if wt[1] == 'VERB' and len(wt[0]) > 4 and not wt[0].endswith("en") and not wt[0].endswith("ing") and (wt[0].endswith("e") or wt[0].endswith("ed"))]
fuzzable_names += [
    "expect",
    "add",
    "report",
    "compare",
    "base",
    "close",
    "offer",
    "price",
    "approve",
    "increase",
    "become",
    "propose",
    "decline",
    "raise",
    "estimate",
    "call",
    "design",
    "acquire",
    "gain",
    "reach",
    "announce",
    "fill",
    "became",
    "include",
    "decid",
    "disclose",
    "agree",
    "fail",
    "complete",
    "raise",
    "trade",
    "continue",
    "include",
    "believe",
    "receive",
    "schedule",
    "indicate",
    "provide",
    "help",
    "need",
    "cause",
    "drop",
    "show",
    "order",
    "chang",
    "launch",
    "reduce",
    "plan",
    "want",
    "follow",
    "trade",
    "improve",
    "issu",
    "involve",
    "reject",
    "increase",
    "turn",
    "barr",
    "earn",
    "consent",
    "total",
    "acquire",
    "require",
    "prefer",
    "produce",
    "introduce",
    "consider",
    "suspend",
    "prove",
    "open",
    "close",
    "boost",
    "list",
    "prepare",
    "allow",
    "approve",
    "wrote",
    "reduce",
    "advance",
    "describe",
    "produce",
    "operate",
    "surge",
    "jump",
    "provide",
    "mature",
    "stop",
    "work",
    "introduce",
    "relate",
    "improve",
    "seem",
    "force",
    "leave",
    "believe",
    "develop",
    "decline",
    "expire",
    "invest",
    "settle",
    "change",
    "contribute",
    "elaborate",
    "refuse",
    "quote",
    "pass",
    "threaten",
    "cause",
    "violate",
    "soar",
    "eliminate",
    "create",
    "replace",
    "argue",
    "elect",
    "complete",
    "issue",
    "register",
    "pursue",
    "combine",
    "start",
    "cover",
    "eliminate",
    "plunge",
    "contain",
    "manag",
    "suggest",
    "appear",
    "discover",
    "oppose",
    "form",
    "limit",
    "force",
    "disgorge",
    "attribut",
    "studi",
    "resign",
    "settl",
    "retire",
    "move",
    "anticipat",
    "decide",
    "prompt",
    "maintain",
    "rang",
    "focus",
    "climb",
    "adjust",
    "award",
    "carri",
    "identif",
    "confirm",
    "match",
    "conclude",
    "sign",
    "adopt",
    "accept",
    "expand",
    "exercise",
    "finish",
    "finance",
    "charge",
    "realize",
    "remain",
    "express",
    "replace",
    "deliver",
    "dump",
    "import",
    "assume",
    "capture",
    "join",
    "releas",
    "lower",
    "exce",
    "determin",
    "locat",
    "appli",
    "complain",
    "trigger",
    "enter",
    "intend",
    "purchas",
    "blam",
    "learn",
    "renew",
    "view",
    "speculat",
    "choose",
    "respond",
    "encourage",
    "dismiss",
    "realiz",
    "serve",
    "associat",
    "slipp",
    "spark",
    "negotiate",
    "pleas",
    "divid",
    "financ",
    "execute",
    "discuss",
    "hir",
    "reflect",
    "determine",
    "market",
    "warn",
    "qualifi",
    "explain",
    "impose",
    "recognize",
    "indicate",
    "point",
    "collect",
    "benefit",
    "attach",
    "compete",
    "incurr",
    "remove",
    "share",
    "request",
    "permit",
    "remove",
    "revive",
    "predict",
    "couple",
    "play",
    "commit",
    "revive",
    "kill",
    "present",
    "deserve",
    "convict",
    "agree",
    "accommodate",
    "surrender",
    "restore",
    "restructur",
    "fear",
    "represent",
    "fund",
    "involve",
    "redeem",
    "resolve",
    "obtain",
    "employ",
    "promote",
    "impose",
    "insist",
    "contact",
    "print",
    "advertise",
    "damage",
    "exercis",
    "auction",
    "disappoint",
    "subordinat",
    "secure",
    "integrat",
    "perform",
    "stepp",
    "regulat",
    "trail",
    "occurr",
    "expell",
    "amount",
    "back",
    "regard",
    "conclude",
    "dominat",
    "push",
    "rumor",
    "respect",
    "specifi",
    "support",
    "abandon",
    "figure",
    "extend",
    "tender",
    "unveil",
    "expose",
    "industrializ",
    "regulate",
    "contract",
    "lift",
    "welcom",
    "squeez",
    "prolong",
    "record",
    "announce",
    "assert",
    "inch",
    "manufacture",
    "describe",
    "calculate",
    "favor",
    "train",
    "institut",
    "concern",
    "accelerat",
    "solve",
    "store",
    "assembl",
    "link",
    "advertis",
    "kick",
    "scrambl",
    "skyrocket",
    "target",
    "crippl",
    "stress",
    "manufactur",
    "provoke",
    "handle",
    "poll",
    "endors",
    "balk",
    "compensate",
    "terminate",
    "operate",
    "admit",
    "attract",
    "feature",
    "devote",
    "triple",
    "concentrat",
    "plead",
    "inspir",
    "defend",
    "treat",
    "violat",
    "enforce",
    "surfac",
    "concentrate",
    "suffer",
    "advis",
    "interview",
    "gauge",
    "measur",
    "hamper",
    "nominat",
    "assure",
    "merge",
    "achieve",
    "retain",
    "chair",
    "relegat",
    "mount",
    "Ask",
    "compil",
    "Guarante",
    "position",
    "lock",
    "roll",
    "drift",
    "Estimat",
    "persuade",
    "survive",
    "Found",
    "chastis",
    "handl",
    "press",
    "sweeten",
    "allocat",
    "criticiz",
    "place",
    "prais",
    "install",
    "weigh",
    "perceiv",
    "remark",
    "moderat",
    "stat",
    "rush",
    "surpris",
    "collaps",
    "licens",
    "disagree",
    "publiciz",
    "pressure",
    "drive",
    "omitt",
    "assum",
    "switch",
    "define",
    "sound",
    "invent",
    "absorb",
    "found",
    "observ",
    "desir",
    "sustain",
    "welcome",
    "load",
    "engag",
    "drove",
    "pegg",
    "compromise",
    "enact",
    "negotiate",
    "result",
    "prove",
    "examine",
    "connect",
    "subscribe",
    "organi",
    "diminish",
    "purchase",
    "answer",
    "orient",
    "control",
    "Post",
    "succeed",
    "rewrite",
    "nominate",
    "discharge",
    "entrust",
    "range",
    "attempt",
    "recover",
    "maximize",
    "engage",
    "obligat",
    "label",
    "refuse",
    "denounce",
    "seize",
    "halt",
    "transform",
    "contribute",
    "tolerate",
    "cool",
    "overcome",
    "caution",
    "claim",
    "discontinu",
    "select",
    "participate",
    "bolster",
    "devise",
    "doubt",
    "write",
    "exchange",
    "narrow",
    "strike",
    "diagnos",
    "classif",
    "outlaw",
    "ventilat",
    "slide",
    "track",
    "lengthen",
    "ensnarl",
    "oversee",
    "renovat",
    "accumulat",
    "underscore",
    "guarante",
    "shore",
    "evaluat",
    "clutter",
    "refile",
    "expedit",
    "disput",
    "refund",
    "scrapp",
    "complicate",
    "exist",
    "Regard",
    "halve",
    "store",
    "adapt",
    "achiev",
    "resume",
    "assist",
    "incorporat",
    "capp",
    "stake",
    "outpac",
    "burn",
    "clobber",
    "alarm",
    "fatten",
    "amend",
    "book",
    "watch",
    "number",
    "whistle",
    "perpetuate",
    "root",
    "publish",
    "abide",
    "ration",
    "host",
    "assign",
    "designat",
    "survey",
    "espouse",
    "strapp",
    "twinn",
    "authoriz",
    "paint",
    "accru",
    "swapp",
    "obsess",
    "Film",
    "jostle",
    "populat",
    "curl",
    "dream",
    "resonate",
    "glamorize",
    "collaborat",
    "enabl",
    "chopp",
    "celebrate",
    "scatter",
    "prosecute",
    "unleash",
    "Compare",
    "superimpos",
    "nurtur",
    "shake",
    "interrogat",
    "clean",
    "knitt",
    "assemble",
    "voice",
    "monopolize",
    "spott",
    "Confront",
    "underline",
    "prosecut",
    "enhanc",
    "depend",
    "inflat",
    "educat",
    "fad",
    "stabb",
    "resolv",
    "usher",
    "struggl",
    "distinguish",
    "prepare",
    "copi",
    "broke",
    "car",
    "crowd",
    "decri",
    "overus",
    "enrag",
    "expung",
    "crank",
    "touch",
    "replicat",
    "devis",
    "replicate",
    "discontinue",
    "recommend",
    "embroil",
    "defuse",
    "judg",
    "polariz",
    "discourage",
    "regenerate",
    "Rekindl",
    "averag",
    "protect",
    "prohibit",
    "initiat",
    "mail",
    "quipp",
    "advocate",
    "appoint",
    "exhibit",
    "empower",
    "manipulate",
    "specialize",
    "summon",
    "apologize",
    "emerge",
    "phase",
    "fabricate",
    "speculate",
    "buoy",
    "convinc",
    "erode",
    "trac",
    "recede",
    "flood",
    "bill",
    "alienat",
    "portray",
    "recycle",
    "service",
    "Develop",
    "confuse",
    "materialize",
    "convert",
    "equipp",
    "depress",
    "enclos",
    "single",
    "zoom",
    "command",
    "exhaust",
    "yield",
    "talk",
    "excit",
    "overpric",
    "expir",
    "postpon",
    "reschedul",
    "evaporat",
    "rebuff",
    "review",
    "clamp",
    "interest",
    "license",
    "patent",
    "stirr",
    "devot",
    "escalat",
    "clarifi",
    "cross",
    "penetrate",
    "guid",
    "milk",
    "generate",
    "double",
    "compet",
    "borrow",
    "computerize",
    "analyze",
    "cultivat",
    "tailor",
    "delete",
    "experience",
    "troubl",
    "institute",
    "reopen",
    "knock",
    "synchronize",
    "aggravat",
    "anger",
    "annoy",
    "attend",
    "evoke",
    "scrape",
    "state",
    "memorize",
    "muffl",
    "stare",
    "advance",
    "fill",
    "sack",
    "entitl",
    "dress",
    "decorat",
    "unsettl",
    "breathe",
    "tank",
    "escap",
    "declare",
    "measure",
    "infring",
    "establish",
    "lack",
    "spoke",
    "afflict",
    "harp",
    "seduce",
    "remind",
    "reprove",
    "deteriorat",
    "codifi",
    "pull",
    "acknowledge",
    "hop",
    "arriv",
    "discard",
    "click",
    "visit",
    "disapprove",
    "defin",
    "disciplin",
    "Reach",
    "rectifi",
    "screw",
    "block",
    "emigrate",
    "imagine",
    "tighten",
    "reap",
    "ascribe",
    "tout",
    "acced",
    "entice",
    "pick",
    "impress",
    "entic",
    "befuddl",
    "possess",
    "skipp",
    "graduat",
    "engineer",
    "inherit",
    "diversifi",
    "provok",
    "reallocate",
    "stripp",
    "reallocat",
    "broaden",
    "instruct",
    "draft",
    "waive",
    "bounce",
    "repair",
    "propose",
    "alter",
    "correct",
    "promis",
    "impli",
    "emphasiz",
    "predispose",
    "compos",
    "quote",
    "robb",
    "bother",
    "chose",
    "participat",
    "Choose",
    "depriv",
    "override",
    "impede",
    "impair",
    "dubb",
    "propagandize",
    "clipp",
    "transcribe",
    "happen",
    "disseminate",
    "preclude",
    "mention",
    "examine",
    "disagre",
    "prescribe",
    "assure",
    "react",
    "advocat",
    "convince",
    "exud",
    "Annualiz",
    "clash",
    "evolv",
    "enjoy",
    "recruit",
    "intimidate",
    "Provid",
    "predicat",
    "constru",
    "emasculate",
    "ensure",
    "wast",
    "disapprov",
    "invite",
    "ratifi",
    "characteriz",
    "excise",
    "sav",
    "mortgag",
    "reclaim",
    "parch",
    "profit",
    "curtail",
    "strengthen",
    "cushion",
    "materializ",
    "flirt",
    "fold",
    "brighten",
    "restor",
    "headlin",
    "hand",
    "beleaguer",
    "disclose",
    "approach",
    "screen",
    "miss",
    "preapprov",
    "test-drive",
    "retrac",
    "account",
    "vest",
    "heat",
    "exacerbat",
    "telephone",
    "wedd",
    "wait",
    "sneak",
    "head",
    "curb",
    "battle",
    "entrench",
    "facilitate",
    "stack",
    "constitute",
    "despise",
    "frighten",
    "manage",
    "juggle",
    "automat",
    "dislike",
    "spook",
    "orchestrat",
    "mint",
    "chase",
    "practic",
    "arise",
    "evolve",
    "implement",
    "decrease",
    "sacrifice",
    "Eliminate",
    "please",
    "advise",
    "grapple",
    "appropriat",
    "stifle",
    "notice",
    "document",
    "migrate",
    "color",
    "Fund",
    "Concern",
    "spurn",
    "overvalu",
    "recoup",
    "hunt",
    "promise",
    "breath",
    "enable",
    "combine",
    "insur",
    "look",
    "redistribute",
    "field",
    "concede",
    "endorse",
    "justifi",
    "structur",
    "downgrad",
    "generat",
    "arrang",
    "overstat",
    "circulat",
    "midsiz",
    "eclipse",
    "stretch",
    "debate",
    "assess",
    "revers",
    "chill",
    "insert",
    "outnumber",
    "surge",
    "direct",
    "stimulat",
    "tempt",
    "overdone",
    "waive",
    "notch",
    "search",
    "calculat",
    "tackle",
    "spackle",
    "dispos",
    "stay",
    "revis",
    "conduct",
    "rank",
    "blurr",
    "compare",
    "topp",
    "outdistanc",
    "relaunch",
    "repric",
    "guarantee",
    "hail",
    "despis",
    "subsidize",
    "appease",
    "co-found",
    "coordinate",
    "heighten",
    "nullifi",
    "puzzl",
    "challenge",
    "notifi",
    "clear",
    "delist",
    "explore",
    "emerg",
    "singl",
    "sagg",
    "grant",
    "confus",
    "complicat",
    "Continu",
    "ignor",
    "secede",
    "accrue",
    "term",
    "stemm",
    "magnifi",
    "reimburs",
    "arrive",
    "firm",
    "falter",
    "sense",
    "distribut",
    "experienc",
    "customiz",
    "deriv",
    "avert",
    "slat",
    "realestate",
    "reorganiz",
    "plagu",
    "bounc",
    "reaffirm",
    "demobilize",
    "brush",
    "sabotage",
    "assassinat",
    "avenge",
    "pledg",
    "defeat",
    "rigg",
    "subpoena",
    "plant",
    "explod",
    "centraliz",
    "fizzl",
    "restructure",
    "mitigate",
    "reserv",
    "batter",
    "induce",
    "investigate",
    "estimate",
    "question",
    "trimm",
    "detail",
    "travel",
    "plugg",
    "fashion",
    "arrest",
    "Absorb",
    "finaliz",
    "lease",
    "dash",
    "sputter",
    "harvest",
    "gyrate",
    "scrounge",
    "alleviate",
    "slow",
    "deplete",
    "relieve",
    "compress",
    "delay",
    "influence",
    "plummet",
    "exceed",
    "coat",
    "scuttle",
    "share",
]

secrets_regex = {
    "artifactory": [
        re.compile(r'(?:\s|=|:|"|^)AKC[a-zA-Z0-9]{10,}'),
        re.compile(r'(?:\s|=|:|"|^)AP[\dABCDEF][a-zA-Z0-9]{8,}'),
    ],
    "aws": [
        re.compile(
            r"(A3T[A-Z0-9]|AKIA|AGPA|AIDA|AROA|AIPA|ANPA|ANVA|ASIA)[A-Z0-9]{16}"
        ),
        re.compile(r"""(?i)aws(.{0,20})?['"][0-9a-zA-Z/+]{40}['"]"""),
        re.compile(
            r"""amzn\.mws\.[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"""
        ),
        re.compile(r"da2-[a-z0-9]{26}"),
        re.compile(r"s3\\.amazonaws\\.com"),
        re.compile(r"ec2-[0-9a-z.\\-_]+\\.compute(-1)?\\.amazonaws\\.com"),
        re.compile(r"[0-9a-z.\\-_]+\\.elb\\.[0-9a-z.\\-_]+\\.amazonaws\\.com"),
        re.compile(r"[0-9a-z.\\-_]+\\.rds\\.amazonaws\\.com"),
        re.compile(r"[0-9a-z.\\-_]+\\.cache\\.amazonaws\\.com"),
        re.compile(r"[0-9a-z.\\-_]+\\.s3-website[0-9a-z.\\-_]+\\.amazonaws\\.com"),
        re.compile(r"[0-9a-z]+\\.execute-api\\.[0-9a-z.\\-_]+\\.amazonaws\\.com"),
    ],
    "github": [
        re.compile(r"""(?i)github.{0,3}(token|api|key).{0,10}?([0-9a-zA-Z]{35,40})""")
    ],
    "slack": [re.compile(r"""xox[baprs]-([0-9a-zA-Z]{10,48})?""")],
    "EC": [re.compile(r"""-----BEGIN EC PRIVATE KEY-----""")],
    "DSA": [re.compile(r"""-----BEGIN DSA PRIVATE KEY-----""")],
    "OPENSSH": [re.compile(r"""-----BEGIN OPENSSH PRIVATE KEY-----""")],
    "RSA": [re.compile(r"""-----BEGIN RSA PRIVATE KEY-----""")],
    "PGP": [re.compile(r"""-----BEGIN PGP PRIVATE KEY-----""")],
    "google": [
        re.compile(r"""AIza[0-9A-Za-z\\-_]{35}"""),
        re.compile(r"""[sS][eE][cC][rR][eE][tT].*['|"][0-9a-zA-Z]{32,45}['|"]"""),
        re.compile(r"""[sS][eE][cC][rR][eE][tT].*['|"][0-9a-zA-Z]{32,45}['|"]"""),
        re.compile(r"""[0-9]+-[0-9A-Za-z_]{32}\\.apps\\.googleusercontent\\.com"""),
    ],
    "heroku": [
        re.compile(
            r"""(?i)heroku(.{0,20})?['"][0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}['"]"""
        )
    ],
    "mailchimp": [
        re.compile(r"""(?i)(mailchimp|mc)(.{0,20})?['"][0-9a-f]{32}-us[0-9]{1,2}['"]""")
    ],
    "mailgun": [re.compile(r"""(?i)(mailgun|mg)(.{0,20})?['"][0-9a-z]{32}['"]""")],
    "slack_webhook": [
        re.compile(
            r"""https://hooks.slack.com/services/T[a-zA-Z0-9_]{8}/B[a-zA-Z0-9_]{8}/[a-zA-Z0-9_]{24}"""
        )
    ],
    "stripe": [re.compile(r"""(?i)stripe(.{0,20})?['"][s|k]k_live_[0-9a-zA-Z]{24}""")],
    "square": [
        re.compile(r"""sq0atp-[0-9A-Za-z\-_]{22}"""),
        re.compile(r"""sq0atp-[0-9A-Za-z\-_]{43}"""),
    ],
    "twilio": [re.compile(r"""(?i)twilio(.{0,20})?['"][0-9a-f]{32}['"]""")],
    "dynatrace": [
        re.compile(r"""dt0[a-zA-Z]{1}[0-9]{2}\.[A-Z0-9]{24}\.[A-Z0-9]{64}""")
    ],
    "url": [
        re.compile(r"""(http(s)?|s3)://"""),
        re.compile(
            r"""[a-zA-Z]{3,10}://[^/\\s:@]{3,20}:[^/\\s:@]{3,20}@.{1,100}["'\\s]"""
        ),
        re.compile(
            r"(ftp|jdbc:mysql)://[-a-zA-Z0-9+&@#/%?=~_|!:,.;]*[-a-zA-Z0-9+&@#/%=~_|]"
        ),
    ],
    "authorization": [
        re.compile(
            r"(authorization)\\s*:\\s*(bearer|token|basic)\\s+[0-9a-z\\.\\-_]{6,}"
        ),
        re.compile(r"eyJ[A-Za-z0-9_/+-]*\.[A-Za-z0-9._/+-]*"),
    ],
    "email": [
        re.compile(r"(?<=mailto:)[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9.-]+")
    ],
    "ip": [
        re.compile(
            r"^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])$"
        )
    ],
}


def is_base64(s):
    try:
        return s.endswith("==") or base64.b64encode(base64.b64decode(s)) == s
    except Exception:
        return False


def decode_base64(s):
    s = s.replace("\n", "")
    if is_base64(s):
        decoded = base64.b64decode(s)
        try:
            return decoded.decode()
        except Exception:
            return str(decoded)
    return s


def is_camel_case(s):
    s = re.sub(r"[*._#%&!\"]", "", s)
    for x in string.digits:
        if x in s:
            return False
    for x in string.punctuation:
        if x in s:
            return False
    return s != s.lower() and s != s.upper() and "_" not in s


def calculate_entropy(data):
    if not data or len(data) < 8:
        return 0
    for text in strings_allowlist:
        if text in data:
            return 0
    entropy = 0.0
    # Remove protocol prefixes which tend to increase false positives
    data = re.sub(r"(file|s3|http(s)?|email|ftp)://", "", data)
    if not data:
        return entropy
    digit_found = False
    punctuation_found = False
    ascii_found = False
    for x in charset:
        p_x = float(data.count(x)) / len(data)
        if p_x > 0:
            if not ascii_found and x in string.ascii_letters:
                ascii_found = True
            if not digit_found and x in string.digits:
                digit_found = True
            if not punctuation_found and x in string.punctuation:
                punctuation_found = True
            entropy += -p_x * math.log(p_x, 256)
    if is_camel_case(data) or data.lower() == data or data.upper() == data:
        return min(0.2, round(entropy, 2))
    if ascii_found and (digit_found or punctuation_found):
        if not punctuation_found:
            return min(0.38, round(entropy, 2))
        return round(entropy, 2)
    else:
        return min(0.4, round(entropy, 2))


def check_secret(data):
    for text in strings_allowlist:
        if text in data:
            return None
    for category, rlist in secrets_regex.items():
        for regex in rlist:
            for match in regex.findall(data):
                return category
    return None


def is_binary_string(content):
    """
    Method to check if the given content is a binary string
    """
    textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})
    return bool(content.translate(None, textchars))


def is_ignored_file(file_name):
    """
    Method to find if the given file can be ignored
    :param file_name: File to compare
    :return: Boolean True if file can be ignored. False otherwise
    """
    if not file_name:
        return False
    file_name = file_name.lower()
    extn = "".join(Path(file_name).suffixes)
    if extn in ignore_files or file_name in ignore_files:
        return True
    for ie in ignore_files:
        if file_name.endswith(ie):
            return True
    return False


def is_exe(src):
    """Detect if the source is a binary file

    :param src: Source path

    :return True if binary file. False otherwise.
    """
    if os.path.isfile(src):
        try:
            return is_binary_string(open(src, "rb").read(1024))
        except Exception:
            return False
    return False


def filter_ignored_dirs(dirs):
    """
    Method to filter directory list to remove ignored directories
    :param dirs: Directories to ignore
    :return: Filtered directory list
    """
    [
        dirs.remove(d)
        for d in list(dirs)
        if d.lower() in ignore_directories or d.startswith(".")
    ]
    return dirs


def find_exe_files(src):
    """
    Method to find files with given extenstion
    """
    result = []
    for root, dirs, files in os.walk(src):
        filter_ignored_dirs(dirs)
        for file in files:
            if is_ignored_file(file):
                continue
            fullPath = os.path.join(root, file)
            if is_exe(fullPath):
                result.append(fullPath)
    return result


def bomstrip(manifest):
    """
    Function to delete UTF-8 BOM character in "string"
    """
    utf8bom = b"\xef\xbb\xbf"
    if manifest[:3] == utf8bom:
        return manifest[3:]
    else:
        return manifest


def parse_pe_manifest(manifest):
    """
    Method to parse xml pe manifest
    :param manifest: String manifest
    :return: dict with flattened keys and values
    """
    try:
        attribs_dict = {}
        root = fromstring(bomstrip(manifest))
        for child in root:
            for ele in child.iter():
                attribs_dict[ele.tag.rpartition("}")[-1]] = ele.attrib
        return attribs_dict
    except Exception as e:
        return None


def is_fuzzable_name(name_str):
    if name_str:
        for n in fuzzable_names:
            if n.lower() in name_str:
                return True
        return False
